/**
  *  \file client/act-planet.cc
  *  \brief Planet Actions
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/actions.h"
#include "client/dlg-cargo.h"
#include "client/dlg-fcode.h"
#include "client/dlg-list.h"
#include "game/fcode.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "ui/commands.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "game/player.h"
#include "game/action.h"

/** Friendly code change, UI action. */
void
doPlanetFCodeChange(GPlanet& planet)
{
    GFCodeList list(GFCodeList::getInstance(), planet);
    WFCodeWindow dialog(_("Change Friendly Code"), list);
    dialog.setFCode(planet.getFCode());
    ui_root->centerWidget(dialog);
    ui_root->add(dialog);
    int rv = ui_root->run();
    ui_root->close(dialog);
    if (rv == cm_Enter) {
        planet.setFCode(dialog.getFCode());
        getDisplayedTurn().doUpdates();
    }
}

/** Cargo transfer to/from planet, UI action.
    \param planet    invoking planet
    \param unload    true if 'u' key, false if 'c' key. These have no
                     difference other than the heading of the target
                     selection window. */
void
doPlanetTransfer(GPlanet& planet, bool unload)
{
    GGameTurn& trn = getDisplayedTurn();

    GObjectList list;
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        GShip& ship = trn.getShip(i);
        if (ship.isExistant() && ship.getOwner() == planet.getOwner() && ship.getPos() == planet.getPos())
            list.addObject(trn.getShipType(), ship);
    }
    if (list.getSize() == 0) {
        messageBox(_("There's none of our ships orbiting this planet."),
                   _("Cargo Transfer"));
        return;
    }
    WObjectSelectionList wlist(0, 15, 320, list);
    GObjectId chosen = wlist.doStandardDialog(unload
                                              ? _("Unload ship...")
                                              : _("Transfer cargo to..."),
                                              0 /* FIXME */);
    if (chosen.isValid())
        doPlanetTransferFor(planet, chosen);
}

/** Planet cargo transfer to ship, UI action.
    \param planet  invoking planet
    \param chosen  target. Must be a ship. */
void
doPlanetTransferFor(GPlanet& planet, GObjectId chosen)
{
    GPlanetTransfer left(planet);
    GShipTransfer   right(*checked_cast<GShip*>(&chosen.getObject()));

    doCargoTransfer(left, right);
    getDisplayedTurn().doUpdates();
}

/** Starbase mission change, UI action. */
void
doBaseChangeMission(GPlanet& planet)
{
    UIStringList data;
    for (int i = 0; i <= 6; ++i)
        data.add(i, GPlanet::getBaseMissionName(i));

    UIStandardListbox lb(0, 7, 300, data);
    lb.scrollToKey(planet.getBaseMission());
    if (lb.doStandardListbox(_("Starbase Mission"), 0 /* FIXME: help id */)) {
        planet.setBaseMission(lb.getKey());
        getDisplayedTurn().doUpdates();
    }
}

/** Starbase shipyard order, UI action.
    \param planet the starbase we're talking about.
    \param pe     triggering event, used for binding the menu. */
void
doBaseFixRecycle(GPlanet& planet, const UIEvent& pe)
{
    GStarbaseFixRecycleAction action(planet);
    UIStringList d;
    if (action.canFix())
        d.add(sya_Fix, _("Fix (repair) a ship"));
    if (action.canRecycle())
        d.add(sya_Recycle, _("Recycle a ship"));
    if (planet.getBaseShipyardOrder() == sya_Recycle)
        d.add(sya_None, _("Cancel `recycle' order"));
    if (planet.getBaseShipyardOrder() == sya_Fix)
        d.add(sya_None, _("Cancel `fix' order"));

    if (d.getCount() == 0) {
        messageBox(_("There is no ship at this starbase which could be fixed or recycled."),
                   _("Shipyard"));
        return;
    }

    UIStandardListbox lb(0, 1, d.getWidth(), d);
    if (!lb.doStandardMenu(pe))
        return;
    switch (int cmd = lb.getKey()) {
     case sya_Recycle:
     case sya_Fix:
        {
            GObjectList list;
            action.enumerateShipsFor(cmd, list);
            list.postprocessUser(true);

            const char* title = (cmd == sya_Recycle ? _("Recycle ship") : _("Fix ship"));
            GObjectId chosen = WObjectSelectionList(0, -1, 300, list).doStandardDialog(title, 0 /* FIXME */);
            if (chosen.isValid())
                action.setAction(cmd, chosen.getId());
        }
        break;
     case sya_None:
        action.setAction(sya_None, 0);
    }
    getDisplayedTurn().doUpdates();
}

