/**
  *  \file game/act-base.cc
  */

#include "game/action.h"
#include "cpluslib/assert.h"

/*********************** GPlanetBuildStarbaseAction **********************/

/** \class GPlanetBuildStarbaseAction
    \brief Building and Scraping Starbases

    This class encapsulates the function to build a starbase, or to
    cancel such an order.

    \inv This action has only one cause-of-error, and that's
    facility-in-use. If there are other error modes, rewrite the ctor and
    dtor. */

/** Constructor.
    \param the_planet    where we want to build
    \param container     pays that bill
    \param want_build    whether we're building or scraping a base.
    \pre want_build != the_planet.isBuildingBase() */
GPlanetBuildStarbaseAction::GPlanetBuildStarbaseAction(GPlanet& the_planet, GCargoContainer& container,
                                                       bool want_build)
    : GCargoCostTransaction(container), the_planet(the_planet),
      want_build(want_build)
{
    ASSERT(want_build != the_planet.isBuildingBase());

    GCargo& h = the_planet.getCargoHandler();
    h.incUseCount();
    if (h.isFacilityInUse(GPlanet::pf_BuildBase)) {
        setError(GError(GError::eFacility, _("This facility is already in use")));
    } else {
        h.acquireFacility(GPlanet::pf_BuildBase);

        int factor = want_build ? +1 : -1;
        setCost(el_Tritanium,  factor * 402);
        setCost(el_Duranium,   factor * 120);
        setCost(el_Molybdenum, factor * 340);
        setCost(el_Money,      factor * 900);
    }
}

GPlanetBuildStarbaseAction::~GPlanetBuildStarbaseAction()
{
    GCargo& h = the_planet.getCargoHandler();
    if (!getError())
        h.releaseFacility(GPlanet::pf_BuildBase);
    h.decUseCount();
}

void
GPlanetBuildStarbaseAction::commit()
{
    ASSERT(isValid());
    the_planet.setBuildBaseFlag(want_build);
    GCargoCostTransaction::commit();
}
