/**
  *  \file game/act-structure.cc
  */

#include "game/action.h"
#include "cpluslib/assert.h"
#include "game/error.h"

/********************** GPlanetBuildStructuresAction *********************/

GPlanetBuildStructuresAction::GPlanetBuildStructuresAction(GPlanet& the_planet,
                                                           GCargoContainer& container)
    : GCargoCostTransaction(container),
      the_planet(the_planet)
{
    the_planet.getCargoHandler().incUseCount();
    if (the_planet.getCargoHandler().isFacilityInUse(GPlanet::pf_Structures)) {
        setError(GError(GError::eFacility, _("This facility is already in use")));
    } else {
        the_planet.getCargoHandler().acquireFacility(GPlanet::pf_Structures);
    }

    for (int i = 0; i < MAX; ++i) {
        TPlanetaryStructure ps = TPlanetaryStructure(i);
        data[ps].order = data[ps].min = the_planet.getStructures(ps);
        data[ps].max = the_planet.getMaxStructures(ps);
    }
}

GPlanetBuildStructuresAction::~GPlanetBuildStructuresAction()
{
    if (!getError())
        the_planet.getCargoHandler().releaseFacility(GPlanet::pf_Structures);
    the_planet.getCargoHandler().decUseCount();
}

void
GPlanetBuildStructuresAction::computeCost()
{
    int nmines = data[str_Mines].order - the_planet.getStructures(str_Mines),
        nfact  = data[str_Factories].order - the_planet.getStructures(str_Factories),
        ndef   = data[str_Defense].order - the_planet.getStructures(str_Defense),
        nsbd   = data[str_BaseDefense].order - the_planet.getStructures(str_BaseDefense);

    setCost(el_Money,    4 * nmines + 3 * nfact + 10 * (ndef + nsbd));
    setCost(el_Supplies, nmines + nfact + ndef);
    setCost(el_Duranium, nsbd);
}

int
GPlanetBuildStructuresAction::addStructures(TPlanetaryStructure ps, int count, bool partial)
{
    if (count > 0) {
        /* add things */
        if (data[ps].order + count > data[ps].max) {
            if (!partial)
                return 0;
            count = data[ps].max - data[ps].order;
        }
    } else {
        /* remove things */
        if (data[ps].order + count < data[ps].min) {
            if (!partial)
                return 0;
            count = data[ps].min - data[ps].order;
        }
    }

    data[ps].order += count;
    computeCost();
    return count;
}

int
GPlanetBuildStructuresAction::addStructuresLimitCash(TPlanetaryStructure ps, int count)
{
    /* don't even try when there is a problem */
    if (getError())
        return 0;

    /* check how much we can add according to the rules */
    int rule_count = addStructures(ps, count, true);
    if (rule_count <= 0)
        return rule_count;

    /* okay, we can add that much, but we don't have the cash. */
    while (!isValid()) {
        addStructures(ps, -1, true);
        rule_count--;
    }
    return rule_count;
}

void
GPlanetBuildStructuresAction::commit()
{
    ASSERT(!getError());
    for (int i = 0; i < MAX; ++i) {
        ASSERT(data[i].order <= data[i].max);
        ASSERT(data[i].order >= data[i].min);
        the_planet.setStructures(TPlanetaryStructure(i), data[i].order);
    }
    GCargoCostTransaction::commit();
}
