/**
  *  \file game/host.h
  *  \brief Knowledge about Host Properties
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_HOST_H
#define PCC_V2_GAME_HOST_H

enum GHostType {
    ht_Unknown,                 ///< Not yet determined.
    ht_Host,                    ///< Tim-Host, any version.
    ht_SRace,                   ///< SRace, any version.
    ht_PHost                    ///< PHost, any version.
};

#define MKVERSION(x,y,z) ((x)*100000+(y)*1000+(z))

/** Knowledge about Host properties. This structure provides a
    namespace for all host-specific knowledge required in our client. */
struct GHost {
    static GHostType  host_type;
    static long       host_version;
    static int        hconfig_size;

    /** Declare that we have a HConfig. */
    static inline void setHaveHConfig(int size)
        {
            if (!host_type) host_type = ht_Host;
            hconfig_size = size;
        }
    /** Declare that we have a PConfig. */
    static inline void
    setHavePConfig()
        { if (!host_type) host_type = ht_PHost; }

    /** Set host version. Use this only when you know the version
        exactly. */
    static inline void setHostVersion(long v) { host_version = v; }

    /** Check whether this host is the 'PHost' family. */
    static inline bool isPHost() { return host_type == ht_PHost; }

    /** Check whether this host is the 'Tim-Host' family. */
    static inline bool isTHost() { return host_type != ht_PHost; }

    /** Check whether this host is SRace. */
    static inline bool isSRace() { return host_type == ht_SRace; }

    static void finish();
    static void clear();

    /** Get ship command argument limit. Many versions have this
        limited to 500, PHost 3.3b raised it to 10000. */
    static inline long getCommandArgumentLimit() {
        if (isPHost() && host_version >= MKVERSION(3,3,2))
            return 10000;
        else
            return 500;
    }

    /** Check whether this host version has Death Rays. */
    static inline bool hasDeathRays() {
        return (isPHost() && host_version >= MKVERSION(4,0,0));
    }

    /** Check whether this host has restrictions in loading high-tech
        torps onto low-tech bases. This is a bug introduced in
        3.22.031, and - given the time Tim spends on Host 3.x - is
        expected to stay for a while. These new hosts do not allow
        you to unload torpedoes onto a starbase if the base doesn't have
        enough tech to build these torps. */
    static inline bool hasHighTechTorpedoBug() {
        return (isTHost() && host_version >= MKVERSION(3,22,31));
    }

    /** Check whether siliconoid natives have desert advantage in this
        host. This applies to Tim-Host, as well as PHost 3.3c+. */
    static inline bool hasSiliconoidDesertAdvantage() {
        return (isTHost() || (isPHost() && host_version >= MKVERSION(3,3,3)));
    }

    /** Check whether this host allows large cargo transfers. A large
        cargo transfer is one where the cargo amount in transit plus
        the ship content exceed the cargo room size. This is allowed
        by most Host versions, but was blocked/made configurable in
        recent THosts. */
    static inline bool hasLargeCargoTransfer() {
        return (isPHost() || (isTHost() && host_version >= MKVERSION(3,22,30)));
    }

    /** Check THost alliance support. This function is only applicable
        to Tim's host, and returns the support level.
        - 0 if no alliances supported
        - 1 if normal 'ff' allies
        - 2 if strong 'FF' allies */
    static inline int getTimAllianceLevel() {
        if (host_version == 0 || host_version >= MKVERSION(3,22,39))
            return 2;
        else if (host_version >= MKVERSION(3,22,7))
            return 1;
        else
            return 0;
    }

    /** Check whether the "Lay mines in" mission automatically fills
        in the minefield owner. When enlarging a foreing minefield,
        some PHost versions require you to set an 'miX' friendly code,
        some do not. */
    static inline bool hasAutomaticMineIdentity() {
        return (isPHost() && host_version >= MKVERSION(3,4,3));
    }

    /** Get post-taxation happiness limit. This is the highest level
        of happiness \em{after taxation happiness effect} at which
        you'll still collect taxes */
    static inline int getPostTaxationHappinessLimit() {
        return isPHost() ? 30 : 31;
    }
};
extern GHost host;

#endif
