/**
  *  \file game/spec-hf.cc
  */

#include <vector>
#include "cpluslib/format.h"
#include "game/player.h"
#include "game/spec.h"
#include "util/console.h"
#include "util/filepars.h"
#include "util/misc.h"
#include "util/nls.h"

/** All known hull functions. \todo This is not freed at program end,
    but I wonder whether anyone cares. */
static std::vector<GHullFunction*> functions;

/** Default special/hull assignments. This vector contains,
    alternating, hull and function numbers. */
static std::vector<int> default_specials;

/** \class GHullFunction
    \brief Description of a Hull Function

    Descriptions of all hull functions are read from a file
    'hullfunc.cc'. The definitions are stored in objects of class
    GHullFunction. This class also provides static methods to obtain
    GHullFunction objects from that definition. */

/** Constructor. Builds a new hull function object with specified name
    and numeric code. */
GHullFunction::GHullFunction(int id, string_t name)
    : name(name), description(), id(id), implies(-1)
{ }

GHullFunction::~GHullFunction()
{ }

/** Get device description. If no description has been set, uses the
    name as description (which is not unreasonable for things like
    "Unclonable"). \see setDescription */
string_t
GHullFunction::getDescription() const
{
    if (description.length())
        return expandRaceNames(description);
    else
        return getName();
}

/** Set device description. This is the string which describes the
    device for the user. */
void
GHullFunction::setDescription(string_t description)
{
    this->description = description;
}

/** Set implied device. Each device 'X' can imply another device 'Y'
    which means that a ship which has both can only do 'X' (or, in
    some sense, a ship with 'X' does everything it could do with 'Y').
    For example, "Unclonable" implies "CloneOnce", because a ship
    which has both is unclonable. "AdvancedBioscan" implies "Bioscan"
    because a ship which has both does "AdvancedBioscan". */
void
GHullFunction::setImpliedDevice(int implies)
{
    this->implies = implies;
}

/** Fetch hull function object by device Id.
    \return pointer to device, null if none defined with that Id. */
GHullFunction*
GHullFunction::getHullFunctionById(int id)
{
    for (std::vector<GHullFunction*>::const_iterator i = functions.begin(); i != functions.end(); ++i)
        if ((*i)->getId() == id)
            return *i;
    return 0;
}

/** Fetch hull function object by name.
    \param name     name of device (case insensitive)
    \param partial  if false, require exact match. If true, accepts
                    devices whose name starts with /name/.
    \return pointer to device, null if none matches. */
GHullFunction*
GHullFunction::getHullFunctionByName(string_t name, bool partial)
{
    GHullFunction* rv = 0;

    name = strLCase(name);
    for (std::vector<GHullFunction*>::const_iterator i = functions.begin(); i != functions.end(); ++i) {
        string_t iname = strLCase((*i)->getName());
        if (iname == name) {
            rv = *i;
            break;
        }
        if (partial && iname.length() > name.length() &&
            iname.compare(0, name.length(), name) == 0)
        {
            rv = *i;
            partial = false;
        }
    }
    return rv;
}

/** Add a hull function. Creates a new hull function object with
    specified Id and name, and makes it available to
    getHullFunctionByName and getHullFunctionById. */
GHullFunction*
GHullFunction::addHullFunction(int id, string_t name)
{
    GHullFunction* hf = new GHullFunction(id, name);
    try {
        functions.push_back(hf);
        return hf;
    }
    catch(...) {
        delete hf;
        throw;
    }
}

/** Clear static data of this module. */
static void
clearFunctions()
{
    for (std::vector<GHullFunction*>::iterator i = functions.begin(); i != functions.end(); ++i) {
        delete *i;
        *i = 0;
    }
    std::vector<GHullFunction*>().swap(functions);
    std::vector<int>().swap(default_specials);
}

namespace {
    /** Parser for 'hullfunc.cc'. */
    class GHullfuncReader : public FileParser {
        /** Last seen hull function. Null if none yet. */
        GHullFunction*  last_function;

        /** If set, the last hull function line was bogus, and we're
            ignoring text until the next one. */
        bool            last_bogus;
     public:
        GHullfuncReader()
            : last_function(0), last_bogus(false) { }
        void process(int line_id, string_t line);
        void error(int line_id, string_t text);
    };
}

/** Shortcut function to display errors. */
void
GHullfuncReader::error(int line_id, string_t text)
{
    console.write(LOG_ERROR,
                  format(_("%s:%d: %s -- line has been ignored"))
                  << getFileName() << line_id << text);
}

void
GHullfuncReader::process(int line_id, string_t line)
{
    line = strTrim(line);

    string_t::size_type pos = line.find_first_of("=,");
    if (pos == string_t::npos) {
        error(line_id, _("Syntax error"));
        return;
    }

    if (line[pos] == ',') {
        /* it's a new function */
        string_t flags, name;
        string_t number = strRTrim(line.substr(0, pos));
        int      real_num;

        line.erase(0, pos+1);
        if (!strSplit(line, flags, name, ",")) {
            error(line_id, _("Syntax error"));
            return;
        }
        name = strTrim(name);

        if (!stringValue(number, real_num) || real_num < 0 || real_num > 0x7FFF) {
            error(line_id, format(_("Invalid device number for `%s'"), name));
            return;
        }

        /* avoid dupes */
        if (GHullFunction::getHullFunctionByName(name)) {
            last_bogus = true;
            error(line_id, format(_("Duplicate definition for hull function with name `%s'"), name));
            return;
        }
        if (GHullFunction::getHullFunctionById(real_num)) {
            last_bogus = true;
            error(line_id, format(_("Duplicate definition for hull function #%d"), real_num));
            return;
        }

        last_function = GHullFunction::addHullFunction(real_num, name);
        last_bogus = false;
    } else {
        /* it's an assignment */
        if (last_bogus)
            return;
        if (!last_function) {
            error(line_id, _("Expected function definition"));
            return;
        }

        string_t name  = strTrim(line.substr(0, pos-1));
        string_t value = strTrim(line.substr(pos+1));
        if (stringMatch("Implies", name)) {
            /* 'i' takes device number or name */
            int  number;
            trimComments(value);
            if (!stringValue(value, number)) {
                if (GHullFunction* hf = GHullFunction::getHullFunctionByName(value)) {
                    number = hf->getId();
                } else {
                    error(line_id, format(_("Unknown hull function `%s'"), value));
                    return;
                }
            }
            if (number != last_function->getId())
                last_function->setImpliedDevice(number);
        } else if (stringMatch("Description", name)) {
            /* 'd' takes string */
            last_function->setDescription(value);
        } else if (stringMatch("Standard", name)) {
            /* 's' takes list of hull numbers */
            trimComments(value);
            do {
                string_t lhs = strTrim(strFirst(value, ","));
                int num;
                if (! stringValue(lhs, num) || num <= 0 || num > MAX_HULLS) {
                    error(line_id, format(_("Invalid hull number `%s'"), lhs));
                    break;
                }
                default_specials.push_back(num);
                default_specials.push_back(last_function->getId());
            } while (strRemove(value, ","));
        }
    }
}

/** Initialize module. This reads 'hullfunc.cc' and sets up all static
    data. */
void
GHullFunction::readHullfuncDefinitions()
{
    clearFunctions();

    GHullfuncReader rdr;
    rdr.parseGameFile("hullfunc.usr");
    rdr.parseGameFile("hullfunc.cc");
}

/** Populate hulls with default specials. */
void
GHullFunction::setDefaultSpecials()
{
    for (std::vector<int>::size_type i = 0; i < default_specials.size(); i += 2)
        if (default_specials[i] <= getNumHulls())
            getHull(default_specials[i]).setSpecialFunction(default_specials[i+1], ~0U, true);
}
