/**
  *  \file game/spec.cc
  *
  *  Specification files. This file contains the classes which
  *  correspond to static game information (hullspec etc.), and
  *  functions to load/access the static specification files.
  *
  *  \todo Management of short names
  *  \todo Free-form field for starship hulls?
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "game/spec.h"
#include "io/streams.h"
#include "util/except.h"
#include "util/nls.h"
#include "game/gamedirs.h"
#include "io/iostr.h"
#include "util/console.h"
#include "util/confpars.h"
#include "util/misc.h"
#include "game/config.h"
#include "cpluslib/format.h"

/** Beam specifications. Indexed 0 to NUM_BEAMS-1 */
std::vector<GBeam> beam_spec;

/** Torpedo specifications. Indexed 0 to NUM_TORPS-1 */
std::vector<GTorpLauncher> torp_spec;

/** Engine specifications. Indexed 0 to NUM_ENGINES-1 */
std::vector<GEngine> eng_spec;

/** Hull specification. Indexed 0 to NNN */
std::vector<GHull*> hull_spec;

/** Truehull. For each player the hulls he can build. */
TTruehull truehull_spec;

/** \class GComponent
    \brief Component of a game item

    This class represents an abstract component of something
    (weapon, structure, ...) */
GComponent::~GComponent()
{ }

/********************************* GBeam *********************************/

/** \class GBeam
    \brief Beam Weapon

    This class represents a beam weapon. */
GBeam::GBeam(int aid, const TBeam& adata)
    : GWeapon(aid), data(adata)
{ }

/***************************** GTorpLauncher *****************************/

/** \class GTorpLauncher
    \brief Torpedo Launcher

    This class represents a torpedo launcher. */
GTorpLauncher::GTorpLauncher(int aid, const TTorpedo& adata)
    : GWeapon(aid), data(adata)
{ }

/******************************** GEngine ********************************/

/** \class GEngine
    \brief Engine

    This class represents a starship engine. */
GEngine::GEngine(int aid, const TEngine& adata)
    : GComponent(aid), data(adata)
{ }

/** Get fuel factor. With the standard fuel consumption formula, this
    is the amount of fuel used by a 100000 kt ship travelling one
    month at the specified speed. The value commonly communicated
    to the user is `getFuelFactor(warp)/(warp*warp)' (fuel per lightyear). */
int
GEngine::getFuelFactor(int warp) const
{
    if (warp == 0)
        return 0;

    ASSERT(warp > 0 && warp < 10);
    return data.fuel[warp-1];
}

/********************************* GHull *********************************/

/** \class GHull
    \brief Starship Hull

    Represents a starship hull. This is information gathered
    together from multiple files: HULLSPEC, HULLFUNC.

    When a ship has a device 'X' which implies 'Y' (see
    GHullFunction::setImpliedDevice), we make sure that every ship
    which receives an 'X' device automatically also has an 'Y' (for
    rule tests). When showing such a ship, and you see a ship which
    has an 'X', the 'Y' should not be displayed. */
GHull::GHull(int aid, const THull& adata)
    : GComponent(aid), data(adata)
{ }

GHull::~GHull()
{ }

TCost
GHull::getCost() const
{
    TCost c = { data.mc, data.t, data.d, data.m };
    return c;
}

/** Set special functions.
    \param hf      Function (hf_xxx)
    \param players Bitmask. 1<<1 == Player 1 can build it, etc.
    \param hull    true: this is a hull-specific abilitiy.
                   false: ship-specific (will be attached to ship when built). */
void
GHull::setSpecialFunction(int hf, int16 players, bool hull)
{
    // we maintain data as follows: players bit 1..11 is which players
    // can access the device, bit 0 is whether it's a hull or
    // ship-specific ability.
    players = (players & ~1) | hull;
    for (std::size_t i = 0; i < specials.getSize(); i += 2) {
        if (specials[i] == hf && (specials[i+1] & ~1) == hull) {
            if (specials[i+1] != players) {
                specials[i+1] |= players;
                goto propagate;
            }
            return;
        }
    }

    int16 data[2];
    data[0] = hf;
    data[1] = players;
    specials.append(data, 2);
 propagate:
    GHullFunction* special = GHullFunction::getHullFunctionById(hf);
    if (special && special->getImpliedDevice() >= 0)
        setSpecialFunction(special->getImpliedDevice(), players, hull);
}

void
GHull::clearSpecialFunctions()
{
    specials.clear();
}

/** Check for special function.
    \param hf     function
    \param player player number
    \param hull   true: query hull-specific ability. Every ship of this
                  type owned by the player has the function.
                  false: query ship-specific ability. Every ship of this
                  type built by the player will have the function. */
bool
GHull::canDoSpecial(int hf, int player, bool hull) const
{
    return getPlayersThatCan(hf, hull) & (1 << player);
}

/** Get player mask.
    \param hf     function
    \param hull   true: query hull-specific abilities; false: ship-specific
    \returns bitmask of all players (1<<N) for whom canDoSpecial() would
    return true. */
int
GHull::getPlayersThatCan(int hf, bool hull) const
{
    for (std::size_t i = 0; i < specials.getSize(); i += 2)
        if (specials[i] == hf && (specials[i+1] & 1) == hull)
            return specials[i+1] & ~1;
    return 0;
}

/******************************** Loading ********************************/

/** Load beamspec file. Throws on error. */
static void
loadBeams()
{
    Ptr<Stream> file = spec_file_dir->openFile("beamspec.dat", Stream::C_READ);

    beam_spec.erase(beam_spec.begin(), beam_spec.end());
    for (int i = 1; i <= NUM_BEAMS; ++i) {
        char buffer[TBeam::size];
        TBeam data;

        if (file->read(buffer, sizeof(buffer)) != sizeof(buffer))
            throw FileFormatException(format(_("Error reading `%s'"), "beamspec.dat"));
        getStructure(buffer, data);
        beam_spec.push_back(GBeam(i, data));
    }
}

/** Load torpspec file. Throws on error. */
static void
loadTorps()
{
    Ptr<Stream> file = spec_file_dir->openFile("torpspec.dat", Stream::C_READ);

    torp_spec.erase(torp_spec.begin(), torp_spec.end());
    for (int i = 1; i <= NUM_TORPS; ++i) {
        char buffer[TTorpedo::size];
        TTorpedo data;

        if (file->read(buffer, sizeof(buffer)) != sizeof(buffer))
            throw FileFormatException(format(_("Error reading `%s'"), "torpspec.dat"));
        getStructure(buffer, data);
        torp_spec.push_back(GTorpLauncher(i, data));
    }
}

/** Load engspec file. Throws on error. */
static void
loadEngines()
{
    Ptr<Stream> file = spec_file_dir->openFile("engspec.dat", Stream::C_READ);

    eng_spec.erase(eng_spec.begin(), eng_spec.end());
    for (int i = 1; i <= NUM_ENGINES; ++i) {
        char buffer[TEngine::size];
        TEngine data;

        if (file->read(buffer, sizeof(buffer)) != sizeof(buffer))
            throw FileFormatException(format(_("Error reading `%s'"), "engspec.dat"));
        getStructure(buffer, data);
        eng_spec.push_back(GEngine(i, data));
    }
}

static void
clearSpecials()
{
    for (int i = 1; i < getNumHulls(); ++i)
        getHull(i).clearSpecialFunctions();
}

namespace {
    /** HULLFUNC.TXT parser. */
    class HullfuncParser : public ConfigParser {
        int chosen_hull;
        int chosen_function;
        bool assign_to_hull;
     public:
        HullfuncParser() : chosen_hull(-1), chosen_function(-1), assign_to_hull(true) { }
        void assignSpecial(int line_id, string_t value, bool byrace);
        void assign(int line_id, string_t name, string_t value, string_t line);
        void error(int line_id, string_t msg);
    };
}

void
HullfuncParser::assign(int line_id, string_t name, string_t value, string_t /*line*/)
{
    if (stringMatch("Initialize", name)) {
        /* "Initialize = Clear" or "Initialize = Default" */
        if (stringMatch("Clear", value)) {
            clearSpecials();
        } else if (stringMatch("Default", value)) {
            clearSpecials();
            GHullFunction::setDefaultSpecials();
        } else {
            error(line_id, _("Invalid argument to `Initialize'"));
        }
    } else if (stringMatch("Hull", name)) {
        /* "Hull = number (Garbage)" or "Hull = Start of name" */
        int h;
        string_t::size_type pos;
        if ((stringValue(value, h, pos)
             || (pos > 0 && stringValue(value.substr(0, pos), h)))
            && (h > 0 && h <= getNumHulls()))
        {
            chosen_hull = h;
        } else {
            value = strUCase(value);
            chosen_hull = -1;
            for (int i = 1; i <= getNumHulls(); ++i) {
                if (strUCase(getHull(i).getName()).compare(0, value.length(), value) == 0) {
                    chosen_hull = h;
                    break;
                }
            }
            if (chosen_hull < 0)
                error(line_id, _("Invalid hull name"));
        }
    } else if (stringMatch("Function", name)) {
        /* "Function = Number (garbage)" or "Function = Name" */
        int h;
        string_t::size_type pos;
        if ((stringValue(value, h, pos) || (pos > 0 && stringValue(value.substr(0, pos), h)))
            && h >= 0)
        {
            chosen_function = h;
        } else if (GHullFunction* hf = GHullFunction::getHullFunctionByName(value, true)) {
            chosen_function = hf->getId();
        } else {
            error(line_id, _("Invalid hull function"));
        }
    } else if (stringMatch("Racesallowed", name)) {
        assignSpecial(line_id, value, true);
    } else if (stringMatch("Playersallowed", name)) {
        assignSpecial(line_id, value, false);
    } else if (stringMatch("Assignto", name)) {
        assign_to_hull = stringMatch("Hull", value);
    } else {
        error(line_id, _("Invalid keyword"));
    }
}

void
HullfuncParser::assignSpecial(int line_id, string_t value, bool byrace)
{
    if (chosen_function < 0)
        error(line_id, _("No function selected for assignment"));
    if (chosen_hull < 0)
        error(line_id, _("No hull selected for assignment"));

    unsigned players = 0;
    while (value.length()) {
        bool neg = false;
        if (value[0] == '-') {
            value.erase(0, 1);
            if (value.length() && value[0] >= '0' && value[0] <= '9') {
                neg = true;
                goto digit;
            }
            players = 0;
        } else if (value[0] == '+') {
            value.erase(0, 1);
            if (value.length() && value[0] >= '0' && value[0] <= '9')
                goto digit;
            players = ~0U;
        } else if (value[0] >= '0' && value[0] <= '9') {
         digit:
            string_t::size_type pos;
            int num;
            if (stringValue(value, num, pos)) {
                value.clear();
            } else {
                if (pos && stringValue(value.substr(0, pos), num))
                    /* okay */;
                else
                    error(line_id, _("Invalid number"));
            }
            if (byrace) {
                for (int i = 1; i <= NUM_PLAYERS; ++i) {
                    if (config.PlayerRace[i-1] == num) {
                        if (neg)
                            players &= ~(1U << i);
                        else
                            players |= 1U << i;
                    }
                }
            } else {
                if (num > 0 && num <= NUM_PLAYERS)
                    if (neg)
                        players &= ~(1U << num);
                    else
                        players |= 1U << num;
                else
                    error(line_id, _("Invalid player number"));
            }
        } else {
            value.erase(0, 1);
        }
    }
    getHull(chosen_hull).setSpecialFunction(chosen_function, players, assign_to_hull);
}

void
HullfuncParser::error(int line_id, string_t msg)
{
    console.write(LOG_WARN, format(_("Line %d: %s"), line_id, msg));
}

/** Load Hull function file. */
static void
loadHullfunc(Ptr<Stream> file, bool in_section)
{
    StandardStream in(file);
    HullfuncParser parser;
    parser.parse(in, "hullfunc", in_section);
}

/** Load hullspec and truehull file. Throws on error.
    \todo Load HULLSPEC.CC(?)
    \pre  configuration has been loaded. */
static void
loadHulls()
{
    Ptr<Stream> file = spec_file_dir->openFile("hullspec.dat", Stream::C_READ);

    hull_spec.erase(hull_spec.begin(), hull_spec.end());
    int nhulls = file->getSize() / THull::size;
    if (nhulls < 10)
        throw FileTooShortException("hullspec.dat");
    if (nhulls > MAX_HULLS)
        nhulls = MAX_HULLS;

    for (int i = 1; i <= nhulls; ++i) {
        char buffer[THull::size];
        THull data;

        if (file->read(buffer, sizeof(buffer)) != sizeof(buffer))
            throw FileFormatException(format(_("Error reading `%s'"), "hullspec.dat"));
        getStructure(buffer, data);
        // FIXME: this is (a) not exception safe and (b) leaks
        hull_spec.push_back(new GHull(i, data));
    }

    file = spec_file_dir->openFile("truehull.dat", Stream::C_READ);

    char buffer[TTruehull::size];
    if (file->read(buffer, sizeof(buffer)) != sizeof(buffer))
        throw FileFormatException(format(_("Error reading `%s'"), "truehull.dat"));
    getStructure(buffer, truehull_spec);

    for (int p = 0; p < NUM_PLAYERS; ++p)
        for (int i = 0; i < NUM_HULLS_PER_PLAYER; ++i)
            if (truehull_spec.hulls[p][i] < 0 || unsigned(truehull_spec.hulls[p][i]) > hull_spec.size())
                throw FileFormatException(format(_("File `%s' is invalid"), "truehull.dat"));

    GHullFunction::readHullfuncDefinitions();
    GHullFunction::setDefaultSpecials();

    if ((file = spec_file_dir->openFileNT("shiplist.txt", Stream::C_READ | Stream::C_TEXT)))
        loadHullfunc(file, false);
    else if ((file = spec_file_dir->openFileNT("hullfunc.txt", Stream::C_READ | Stream::C_TEXT)))
        loadHullfunc(file, true);
}

/**************************** Public Interface ***************************/

/** Load Specification Files.
    \pre initGameGlobals() has been called.
    \pre Configuration was loaded. */
void
initSpecification()
{
    loadTorps();
    loadBeams();
    loadEngines();
    loadHulls();
}

/** Get Truehull slot. Returns the position of hull hid in player
    pid's Truehull list, [1, NUM_HULLS_PER_PLAYER]. Returns 0 if he can
    not build that hull. */
int
getTruehullSlot(int pid, int hid)
{
    for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i)
        if (getTruehull(pid, i) == hid)
            return i;
    return 0;
}
