/**
  *  \file int/bytecode.cc
  *  \brief Bytecode
  *
  *  The main reason behind byte-compilation is correctness, not speed.
  *  Of course, nobody will complain when we get faster :)
  *
  *  The main problem is that certain errors will not be found by the
  *  PCC 1.x interpreter. Compilation would solve those. The other
  *  problem is that complete byte-compilation is not possible due to
  *  the simple syntax of the language: a line containing just `foo+1'
  *  could mean `compute the sum of `foo' and `1' and discard the
  *  result', or `call the subroutine `foo' with argument `+1''. We do
  *  not know whether a subroutine `foo' will be defined by the time
  *  this is executed. Hence we only compile built-in commands with
  *  special syntax (among those, the control structures that cause
  *  trouble in PCC 1.x), and handle the rest with `Eval' instructions.
  *
  *  This module contains routines to build up BCOs. Management of
  *  BCOs (which are subroutines) is in int/subr.cc, execution is in
  *  int/exec.cc.
  *
  *  \todo Calling a user-defined subroutine is currently handled
  *  symbolically. For example, "foo \<arg>" is translated as
  *  \verbatim
  *         <arg>
  *         push 1
  *         invokeuser "foo"
  *  \endverbatim
  *  This makes trouble if \<arg> suspends and the re-loaded VM does
  *  not define a subroutine "foo". Another alternative would be to
  *  compile it as
  *  \verbatim
  *         <arg>
  *         push 1
  *         invokeuser <subr-id-of-"foo">
  *  \endverbatim
  *  and store "foo" in the VM image.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#include <iostream>
#include <cstring>
#include <map>
#include "int/bytecode.h"
#include "cpluslib/assert.h"
#include "cpluslib/misc.h"
#include "int/exec.h"

/** Allocate a new BCO. BCOs must be heap-allocated. */
IntByteCodeObject&
IntByteCodeObject::allocBCO()
{
    return *new IntByteCodeObject();
}

/*************************** IntByteCodeObject ***************************/

IntByteCodeObject::IntByteCodeObject()
    : min_args(0), max_args(0)
{ }

IntByteCodeObject::~IntByteCodeObject()
{
    for (index_t i = 0; i < opcodes.size(); ++i)
        if (getOpcodeFromCode(opcodes[i]) == op_DefineSub)
            if (IntSubroutine* s = getSubrByIndex(getArgFromCode(opcodes[i])))
                s->decUseCount();
    // ASSERT(ref_count == 0);
}

/** Add one op-code. */
void
IntByteCodeObject::add(uint32 code)
{
    opcodes.push_back(code);
    if (getOpcodeFromCode(code) == op_DefineSub)
        getSubrByIndex(getArgFromCode(code))->incUseCount();
}

/** Compile code to push an integer literal. */
void
IntByteCodeObject::pushInt(int32 i)
{
    uint32 n = i;               // ISO C++ guarantees 2's complement expansion
    if (i > -0x7FFFFF && i < 0x7FFFFF) {
        add(mkCode(op_PushInt, n));
    } else {
        add(mkCode(op_PushInt,   n & 0xFFFFFF));
        add(mkCode(op_PushIntHi, (n >> 24) & 255));
    }
}

/** Compile code to push a floating-point literal. */
void
IntByteCodeObject::pushFloat(double d)
{
    add(mkCode(op_PushFloat, allocFloat(d)));
}

/** Compile code to push a string literal. */
void
IntByteCodeObject::pushStr(const string_t& str)
{
    add(mkCode(op_PushString, allocStr(str)));
}

/** Get handle for a floating-point value. To access it, pass the
    handle as second argument to mkCode(). */
IntByteCodeObject::index_t
IntByteCodeObject::allocFloat(double d)
{
    float_table.push_back(d);
    return float_table.size() - 1;
}

/** Get handle for a string value. To access it, pass the
    handle as second argument to mkCode(). */
IntByteCodeObject::index_t
IntByteCodeObject::allocStr(const string_t& str)
{
    string_table.push_back(str);
    return string_table.size() - 1;
}

/** Get handle for a label. The label remains undefined. To define it,
    use setLabel(). To access (jump to) it, pass the handle as second
    argument to mkCode(). */
IntByteCodeObject::label_t
IntByteCodeObject::allocLabel()
{
    label_table.push_back(-1);
    return label_table.size() - 1;
}

/** Place a label at the current end of the BCO. */
void
IntByteCodeObject::setLabel(label_t l)
{
    ASSERT(l < label_table.size());
    ASSERT(label_table[l] == uint32(-1));
    label_table[l] = opcodes.size();
}

/** Add one argument to this user-defined subroutine.
    \param name       name of argument
    \param mandatory  true iff mandatory argument
    \pre if this has once been called with mandatory=false, all
    further calls must have mandatory=false, too. */
void
IntByteCodeObject::addArgument(string_t name, bool mandatory)
{
    ASSERT(max_args == arg_names.size());
    ASSERT(min_args == max_args || !mandatory);

    arg_names.push_back(name);
    ++max_args;
    if (mandatory)
        ++min_args;
}

static const char*const insn_def[] = {
    "call1 N",
    "call2 N",
    "call3 N",
    "callv N",

    "invokeu S",
    "invokec S",

    "pushn",
    "pushb N",
    "pushi N",
    "pushf F",
    "pushs S",
    "pushih N",
    "pop N",
    "dup N",
    "dupc N",

    "qload S",
    "qstore S",
    "uload S",
    "ustore S",

    "apply N",
    "astore N",

    "jnever L",
    "jt L",
    "jf L",
    "jne L",
    "je L",
    "jte L",
    "jfe L",
    "j L",
    "jneverp L",
    "jtp L",
    "jfp L",
    "jnep L",
    "jep L",
    "jtep L",
    "jfep L",
    "jp L",

    "eval S",
    "setline N",

    "pushcontext",
    "popcontext N",
    "currentcontext",
    "firstcontext",
    "nextcontext",

    "try L",
    "endtry",
    "throw",

    "definesub N",

    "dimlocal S",
    "dimstatic S",
    "dimshared S",

    "return",

    "stop N"
};
CTASSERT(op_LASTOPCODE+1 == countof(insn_def));

/** Display contents (bytecode) of this BCO on /out/. */
void
IntByteCodeObject::dump(std::ostream& out)
{
    for (std::size_t i = 0; i < opcodes.size(); ++i) {
        for (std::size_t n = 0; n < label_table.size(); ++n)
            if (label_table[n] == i)
                out << "L" << n << ":\n";
        const char* p = insn_def[getOpcodeFromCode(opcodes[i])];
        uint32 arg = getArgFromCode(opcodes[i]);
        const char* q = strchr(p, ' ');
        if (q) {
            out << "\t" << string_t(p, q-p) << "\t";
            switch (q[1]) {
             case 'N':
                out << arg << "\n";
                break;
             case 'S':
                out << "\"" << string_table[arg] << "\" @ " << arg << "\n";
                break;
             case 'F':
                out << float_table[arg] << " @ " << arg << "\n";
                break;
             case 'L':
                out << "L" << arg << "\n";
                break;
             default:
                FAIL();
            }
        } else
            out << "\t" << p << "\n";
    }
    for (std::size_t n = 0; n < label_table.size(); ++n)
        if (label_table[n] == opcodes.size())
            out << "L" << n << ":\n";
}

/** Schedule this BCO for execution in the specified context. */
void
IntByteCodeObject::run(int nargs, string_t name, IntExecutionContext& xc)
{
//     std::cout << "invoking BCO " << getId()
//               << " with " << nargs
//               << " args (need " << min_args << ", max. " << max_args
//               << "), bytecode is\n";
//     dump(std::cout);

    if (nargs < min_args) {
        xc.doException("Too few arguments in call to `" + name + "'");
        return;
    }
    if (nargs > max_args) {
        xc.doException("Too many arguments in call to `" + name + "'");
        return;
    }

    Ptr<IntFrame> f = new IntFrame(*this, xc.getControlStack().size());
    for (int i = nargs; i > 0; --i)
        f->local_values.store(i-1, xc.popV());
    xc.pushFrame(f);
}
