/**
  *  \file ui/skin.cc
  *  \brief Skinning and Color Scheme Management
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/skin.h"
#include "cpluslib/assert.h"
#include "cpluslib/misc.h"
#include "gfx/gfx.h"
#include "gfx/font.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"

/******************************* Color Sets ******************************/

/** Colors used to display something on a light-gray background.
    Examples include a light-gray window, or a standard list box
    entry. */
UIColorSet GRAY_COLOR_SET = {
    COLOR_BLACK,                // static
    COLOR_GREENBLACK,           // green
    COLOR_BRIGHTORANGE,         // yellow
    COLOR_RED,                  // red
    COLOR_WHITE,                // white
    COLOR_BLACK,                // contrast
    COLOR_BLUEGRAY,             // input
    COLOR_BLUEBLACK,            // blue
    COLOR_DARK,                 // faded
    COLOR_GRAY,                 // background
    { }
};

/** Colors used to display something on a dark-gray background.
    Examples include a dark-gray window, or a control screen. */
UIColorSet DARK_COLOR_SET = {
    COLOR_GRAY,                 // static
    COLOR_GREEN,                // green
    COLOR_YELLOW,               // yellow
    COLOR_RED,                  // red
    COLOR_WHITE,                // white
    COLOR_WHITE,                // contrast
    COLOR_YELLOW,               // input
    COLOR_BLUEGRAY,             // blue      // ???
    COLOR_GRAYSCALE+4,          // faded     // ???
    COLOR_DARK,                 // background
    { }
};

/** Colors used to display something on a black background. Examples
    include the selection bar on a list box. */
UIColorSet BLACK_COLOR_SET = {
    COLOR_WHITE,                // static
    COLOR_GREEN,                // green
    COLOR_DARKYELLOWSCALE+6,    // yellow (0xb6 in PCC 1)
    COLOR_RED,                  // red       // FIXME: darker?
    COLOR_WHITE,                // white
    COLOR_WHITE,                // contrast
    COLOR_GRAY,                 // input
    COLOR_BLUEGRAY,             // blue
    COLOR_DARK,                 // faded
    COLOR_BLACK,                // background
    { }
};

/** Colors used to display something bright on a light-gray
    background. This is used for the selection of a listbox when the
    listbox doesn't have focus. */
UIColorSet BRIGHT_GRAY_COLOR_SET = {
    COLOR_WHITE,                // static
    COLOR_GREEN,                // green
    COLOR_YELLOW,               // yellow
    COLOR_RED,                  // red
    COLOR_WHITE,                // white
    COLOR_WHITE,                // contrast
    COLOR_GRAY,                 // input
    COLOR_BLUEGRAY,             // blue
    COLOR_DARK,                 // faded
    COLOR_GRAY,                 // background
    { }
};

/********************************* UISkin ********************************/

/** \class UISkin
    \brief A widget skin

    Well-behaved widgets can use a "skin" to adapt to their
    environment. Skins essentially provide a means to draw the
    background of a widget, and to draw readable text on it. */

UISkin::~UISkin()
{ }

UISkin&
UISkin::getDefaultSkin()
{
    static UIDefaultSkin d;
    return d;
}

void
UISkin::startDraw()
{
    parm.just_x = parm.just_y = 0;
    parm.line_thickness = 1;
    parm.line_style = uint8(~0U);
    parm.fill_pattern = SOLID_FILL;
    parm.alpha = GFX_ALPHA_OPAQUE;
    solid = true;
}

void
UISkin::setColor(Color color)
{
    parm.color = standard_colors[getColors().table[color]];
}

/****************************** Default Skin *****************************/

UIDefaultSkin::UIDefaultSkin()
{ }

UIDefaultSkin::~UIDefaultSkin()
{ }

const UIColorSet&
UIDefaultSkin::getColors() const
{
    return DARK_COLOR_SET;
}

void
UIDefaultSkin::outText(GfxCanvas& can,  int f, int x, int y, string_t text)
{
    parm.font = fonts[f];
    parm.bgcolor = isSolid() ? standard_colors[COLOR_GRAY] : GFX_TRANSPARENT;
    ::outText(can, parm, x, y, text);
}

void
UIDefaultSkin::outTextF(GfxCanvas& can, int f, int x, int y, int lim, string_t text)
{
    parm.font = fonts[f];
    parm.bgcolor = isSolid() ? standard_colors[COLOR_GRAY] : GFX_TRANSPARENT;
    ::outTextF(can, parm, x, y, lim, text);
}

void
UIDefaultSkin::drawBackground(GfxCanvas& can, GfxRect r)
{
    drawSolidBar(can, r, standard_colors[COLOR_GRAY]);
}
